/*****************************************************************************
    This file is part of MP3ext.
    Copyright (C) 1999  Michael Mutschler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    email: michael@mutschler.de
******************************************************************************/
/*
 * ID3V2.c
 *
 * This file includes functions to access the ID3V2 header
 *
 */

#include "ID3V1.h"
#include "ID3V2Errors.h"
#include <stdio.h>
#include <string.h>
#ifdef WIN32
#include <windows.h>

//#ifdef _DEBUG
//#define _CRTDBG_MAP_ALLOC
//#endif
//#include <crtdbg.h>
#endif

#include <tchar.h>
#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#endif


/* local functions */
#define IsID3Tag(s) (s[0] == 'T' && s[1] == 'A' && s[2] == 'G')
static FILE *SeekEnd(LPCTSTR file);
static int RemoveID3(LPCTSTR file);
unsigned char GenreStr2Int(LPCTSTR str);
LPCTSTR GenreInt2Str(unsigned char i);
char *TrimString(char *s, int maxlen);


/* ******************************************************************** */
/* this function reads the ID3V1 info from a given file-pointer         */
CGLOBAL int ReadID3V1Info(LPCTSTR file, struct ID3V1Info *id3i)
{
	FILE *fh;
	int err=1;
	size_t readbytes;

	/* open file, and seek to the beginning of a possible ID3-Tag */
	if ((fh = _tfopen(file, _T("rb"))) && (fseek(fh, -ID3TAG_LENGTH, SEEK_END) == 0))
	{
		readbytes = fread(id3i->Tag, 1, ID3TAG_LENGTH, fh);
		fclose(fh);

		/* check, if we have a valid ID3-Tag */
		if ((readbytes == ID3TAG_LENGTH) && (IsID3Tag(id3i->Tag)))
		{
			// trim the strings
			#define TRIM(field) \
				TrimString((char *)id3i->field, sizeof(id3i->field))
			TRIM(SongName);
			TRIM(Artist);
			TRIM(Album);
			TRIM(Year);
			TRIM(Comment);
			#undef TRIM
			
			id3i->Genre = id3i->Genre + 1;

			/* now the track-field */
			if (id3i->Comment[28] == '\0')
				id3i->Track = id3i->Comment[29];
			else
				id3i->Track = 0;

			err = 0;
		}
		else
		{	/* clear the structure */
			memset(id3i, 0, sizeof(struct ID3V1Info));
		}
	}
	return(err);
}

/* ******************************************************************** */
/* write an ID3v1 structure in the file
 *
 * flags indicate:
 * ID3V1_WRITE_REMOVE  - remove the ID3-Tag if possible
 * ID3V1_WRITE_REPLACE - write only, if there exists one
 * ID3V1_WRITE_NORMAL  - overwrite if necessary, else add a new one
 *
 * if id3i is NULL, then remove the Tag always!
 */
CGLOBAL int WriteID3V1Info(LPCTSTR file, struct ID3V1Info *id3i, int flags)
{
	FILE *fh;
	int err=1;
	
	if (flags == ID3V1_WRITE_REMOVE || id3i == NULL)	// remove if none exists
	{
		return(RemoveID3(file));
	}

	/* set the TAG */
	id3i->Tag[0] = 'T';
	id3i->Tag[1] = 'A';
	id3i->Tag[2] = 'G';

	/* now prepare the structure for writing*/
	/* fill the strings with zeros after the end */
		// trim the strings
		#define TRIM(field) \
			TrimString((char *)id3i->field, sizeof(id3i->field))
		TRIM(SongName);
		TRIM(Artist);
		TRIM(Album);
		TRIM(Year);
		TRIM(Comment);
		#undef TRIM
	
	/* now the Genre: */
	id3i->Genre = id3i->Genre - 1;

	/* now the track */
	if (id3i->Track != 0)		// write a track, if there exists one
	{
		id3i->Comment[28] = 0;
		id3i->Comment[29] = (unsigned char)id3i->Track;
	}

	/* structure is now prepared for writing */
	err = 0;
	fh = SeekEnd(file);

	if (fh == 0)
		return(ID3V2_ERR_WRITE);
	else
	{
		if (flags == ID3V1_WRITE_NORMAL ||	/* write tag if normal mode */
			 !feof(fh))						/* or not at eof -> there is another ID3V1-Tag in the file */
		{
			fwrite(id3i->Tag, 1, ID3TAG_LENGTH, fh);
		}
		fclose(fh);
	}

	return(err);
}

/* ******************************************************************** */
/* Return TRUE if ID3V1 TAG exist in this file */
CGLOBAL BOOL ExistID3Tagv1(LPCTSTR file)
{
	FILE *fh;
	char head[3];
	size_t readbytes;
	BOOL bExistID3v1 = FALSE;

	if ((fh=_tfopen(file, _T("rb"))) == NULL)
		return(FALSE);

	// ok, file is open, now look if there is a tag
	if ((fseek(fh, -ID3TAG_LENGTH, SEEK_END) == 0) &&
		 (readbytes = fread(head, 1, sizeof(head), fh)) &&
		 readbytes == sizeof(head) &&
		 IsID3Tag(head)
		)
	{
		bExistID3v1 = TRUE;
	}
	
	fclose(fh);

	return(bExistID3v1);
}


/* ******************************************************************** */
// if maxlen = -1 then use the complete string
char *TrimString(char *s, int maxlen)
{
	int i;
	int len=strlen(s);

	if (s == NULL) return(s);	// just to be sure we have a string
	len=strlen(s);
	if (maxlen != -1 && maxlen<len) len = maxlen;	// make sure we don not exceed our limit

	// first, the beginning of the string:
	for (i=0; i<len; i++) if (!isspace(s[i])) break;

	if (i>0)	// if the string hasn't any leading spaces, skip this part
	{
		if (i==len) {*s='\0'; return(s);}		// a blank string...

		// new length of the string
		len -= i;

		// move the string to the beginning:
		memmove(s, s+i, len);
		s[len] = '\0';		// new end
	}

	len--;
	// now check the end
	while (len>=0) {
		if (isspace(s[len])) s[len] = '\0';
		else break;	// until the first char occures
		len--;
	}
	return(s);
}

/* ******************************************************************** */
static FILE *SeekEnd(LPCTSTR file)
{
	FILE *fh;
	char head[3];
	size_t readbytes;

	if ((fh=_tfopen(file, _T("r+b"))) == NULL)
		return(NULL);

	// ok, file is open, now look if there is a tag
	if ((fseek(fh, -ID3TAG_LENGTH, SEEK_END) == 0) &&
		 (readbytes = fread(head, 1, sizeof(head), fh)) &&
		 readbytes == sizeof(head) &&
		 IsID3Tag(head)
		)
	{
		// there already exists an id3-Tag
		fseek(fh, -ID3TAG_LENGTH, SEEK_END);
	}
	else fseek(fh, 0, SEEK_END);

	return(fh);
}

/* ******************************************************************** */
// return TRUE if error
static int RemoveID3(LPCTSTR file)
{
	FILE *fh;
	if (!(fh = SeekEnd(file))) return(1);	// error
	/* now set the eof-marker at this position, and close the file
	 * but how do you do it in ANSI-C?
	 */
#ifdef WIN32
	{
		long offset;		// remeber the offset in the file
	
		HANDLE wfh;

		offset = ftell(fh);
		fclose(fh);
		if ((wfh = CreateFile(file,
								GENERIC_WRITE|GENERIC_READ,
								FILE_SHARE_READ|FILE_SHARE_WRITE,
								NULL,
								OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL|FILE_FLAG_RANDOM_ACCESS,
								NULL))
			== INVALID_HANDLE_VALUE)
		return(1);		// error
		SetFilePointer(wfh, offset, NULL, FILE_BEGIN);
		SetEndOfFile(wfh);
		CloseHandle(wfh);
	}
#else
	// ret = SetEndOfFile(fh);
	fclose(fh);
#endif
	return(0);
}

/* ******************************************************************** */
unsigned char GenreStr2Int(LPCTSTR str)
{
	unsigned char genre = -1;
	if (str != NULL)
	{
		while (ID3_Genres[++genre])
			if (_tcsicmp(str, ID3_Genres[genre]) == 0)
				return(genre);
		// if nothing matched, return the default
	}
	return(ID3_DEFGENRE);
}


/* ******************************************************************** */
LPCTSTR GenreInt2Str(unsigned char genre)
{
	return (ID3_Genres[((genre>= 0 && genre <= ID3TAG_MAX_GENRE) ? genre : ID3_DEFGENRE)]);
}



LPTSTR ID3_Genres[ID3TAG_MAX_GENRE+2] = {
		 _T("???"),			// additional
		 _T("Blues"),
		 _T("Classic Rock"),
		 _T("Country"),
		 _T("Dance"),
		 _T("Disco"),
		 _T("Funk"),
		 _T("Grunge"),
		 _T("Hip-Hop"),
		 _T("Jazz"),
		 _T("Metal"),
		 _T("New Age"),
		 _T("Oldies"),
		 _T("Other"),
		 _T("Pop"),
		 _T("R&B"),
		 _T("Rap"),
		 _T("Reggae"),
		 _T("Rock"),
		 _T("Techno"),
		 _T("Industrial"),
		 _T("Alternative"),
		 _T("Ska"),
		 _T("Death Metal"),
		 _T("Pranks"),
		 _T("Soundtrack"),
		 _T("Euro-Techno"),
		 _T("Ambient"),
		 _T("Trip-Hop"),
		 _T("Vocal"),
		 _T("Jazz+Funk"),
		 _T("Fusion"),
		 _T("Trance"),
		 _T("Classical"),
		 _T("Instrumental"),
		 _T("Acid"),
		 _T("House"),
		 _T("Game"),
		 _T("Sound Clip"),
		 _T("Gospel"),
		 _T("Noise"),
		 _T("AlternRock"),
		 _T("Bass"),
		 _T("Soul"),
		 _T("Punk"),
		 _T("Space"),
		 _T("Meditative"),
		 _T("Instrumental Pop"),
		 _T("Instrumental Rock"),
		 _T("Ethnic"),
		 _T("Gothic"),
		 _T("Darkwave"),
		 _T("Techno-Industrial"),
		 _T("Electronic"),
		 _T("Pop-Folk"),
		 _T("Eurodance"),
		 _T("Dream"),
		 _T("Southern Rock"),
		 _T("Comedy"),
		 _T("Cult"),
		 _T("Gangsta"),
		 _T("Top 40"),
		 _T("Christian Rap"),
		 _T("Pop/Funk"),
		 _T("Jungle"),
		 _T("Native American"),
		 _T("Cabaret"),
		 _T("New Wave"),
		 _T("Psychedelic"),
		 _T("Rave"),
		 _T("Showtunes"),
		 _T("Trailer"),
		 _T("Lo-Fi"),
		 _T("Tribal"),
		 _T("Acid Punk"),
		 _T("Acid Jazz"),
		 _T("Polka"),
		 _T("Retro"),
		 _T("Musical"),
		 _T("Rock & Roll"),
		 _T("Hard Rock"),
		 _T("Folk"),
		 _T("Folk/Rock"),
		 _T("National Folk"),
		 _T("Swing"),
		 _T("Fast Fusion"),
		 _T("Bebob"),
		 _T("Latin"),
		 _T("Revival"),
		 _T("Celtic"),
		 _T("Bluegrass"),
		 _T("Avantgarde"),
		 _T("Gothic Rock"),
		 _T("Progressive Rock"),
		 _T("Psychedelic Rock"),
		 _T("Symphonic Rock"),
		 _T("Slow Rock"),
		 _T("Big Band"),
		 _T("Chorus"),
		 _T("Easy Listening"),
		 _T("Acoustic"),
		 _T("Humour"),
		 _T("Speech"),
		 _T("Chanson"),
		 _T("Opera"),
		 _T("Chamber Music"),
		 _T("Sonata"),
		 _T("Symphony"),
		 _T("Booty Bass"),
		 _T("Primus"),
		 _T("Porn Groove"),
		 _T("Satire"),
		 _T("Slow Jam"),
		 _T("Club"),
		 _T("Tango"),
		 _T("Samba"),
		 _T("Folklore"),
		 _T("Ballad"),
		 _T("Power Ballad"),
		 _T("Rhythmic Soul"),
		 _T("Freestyle"),
		 _T("Duet"),
		 _T("Punk Rock"),
		 _T("Drum Solo"),
		 _T("Acapella"),
		 _T("Euro-House"),
		 _T("Dance Hall"),
		 _T("Goa"),
		 _T("Drum & Bass"),
		 _T("Club-House"),
		 _T("Hardcore"),
		 _T("Terror"),
		 _T("Indie"),
		 _T("BritPop"),
		 _T("Negerpunk"),
		 _T("Polsk Punk"),
		 _T("Beat"),
		 _T("Christian Gangsta Rap"),
		 _T("Heavy Metal"),
		 _T("Black Metal"),
		 _T("Crossover"),
		 _T("Contemporary Christian"),
		 _T("Christian Rock"),
		 _T("Merengue"),
		 _T("Salsa"),
		 _T("Thrash Metal"),
		  NULL
};			// additional
