/*****************************************************************************
    This file is part of MP3ext.
    Copyright (C) 1999  Michael Mutschler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    email: michael@mutschler.de
******************************************************************************/
/*
 * ID3V2.c
 *
 * This file includes functions to access the ID3V2 header
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <malloc.h>

#ifdef WIN32
#include <windows.h>

//#ifdef _DEBUG
//#define _CRTDBG_MAP_ALLOC
//#endif
//#include <crtdbg.h>
#endif
#include <tchar.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#endif


#include "zlib.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/* visual C requires this undef otherwise a change in ID3V2fields.h would
 * generate compiler-errors. It seems it doesn't handle multiple includes
 * correct!!! */
#undef ID3V2_CREATE_ERROR_ARRAY
#include "ID3V2.h"
/* Include the Stringtable for the errors */
#define   ID3V2_CREATE_ERROR_ARRAY 1
#include "ID3V2Errors.h"
#undef    ID3V2_CREATE_ERROR_ARRAY
#include "ID3V2fields.h"

/**************************************************************
 * local defined structures
 **************************************************************/

struct ID3V2_Header
{
	char Ident[3];
	char Version;
	char Revision;
	unsigned char Flags;
	unsigned char Size[4];
};
#define IDENTSTRING "ID3"


struct FrameHeader2
{
	char Ident[ID3V22_FRAMEIDLEN];
	unsigned char Size[3];
};

struct FrameHeader3
{
	char Ident[ID3V2_FRAMEIDLEN];
	unsigned char Size[4];
	unsigned char flags[2];
};

// macro for getting the ident-size
#define TAGSIZE (ih->ID3v2Version==2?ID3V22_FRAMEIDLEN:ID3V2_FRAMEIDLEN)

/* this is the buffersize for copying the data */
#define BUFSIZE (512*1024)

/**************************************************************
 * local function definitions
 **************************************************************/

static int OpenRead (LPID3V2Handle ih);
static int OpenWrite(LPID3V2Handle ih);
static int CloseWrite(LPID3V2Handle ih);
static int ReadFrame(LPID3V2Handle ih);
static int WriteFrame(LPID3V2Handle ih);
static int ReadID3V2Header(FILE *fh, struct ID3V2_Header *id3header, int *ID3V2size);
static int decodeFrame(LPID3V2Handle ih);
static int compressFrame(LPID3V2Handle ih);

/* This is the global function to read/write the MP3Info */
CGLOBAL LPID3V2Handle ID3V2OpenFile(LPCTSTR file, int mode, int *ErrorCode, int data)
{
	LPID3V2Handle ih;


	*ErrorCode = ID3V2_ERR_NOMEM;
	if (ih=(LPID3V2Handle)calloc(1, sizeof(ID3V2Handle)))
	{
		/* store some data */
		ih->Mode = mode;	// store the mode here
		ih->oldfh = NULL;
		ih->PaddingSize = 256;	// by default, padding is 256 bytes
		ih->MinCompressLength = -1;	// no comression
		_tcsncpy((LPTSTR)(ih->FileName), file, MAX_FILE);	// copy the filename

		/* set the length, if one is known */
		if (mode & ID3V2_MODE_HAVELEN)
			ih->ID3v2Length = data;

		/* now determine, if we have a read or write-request */
		if (mode & ID3V2_MODE_WRITE)
			*ErrorCode = OpenWrite(ih);
		else
			*ErrorCode = OpenRead(ih);

		/* if there was an error, close the free the handle */
		if (*ErrorCode != ID3V2_ERR_OK)
		{
			ID3v2CloseHandle(ih);
			ih = NULL;
		}
	}
	return(ih);
}

/* process the next frame */
CGLOBAL int ID3v2NextFrame(LPID3V2Handle ih)
{
	assert(ih);
	if (ih->Mode & ID3V2_MODE_WRITE)	return(WriteFrame(ih));
	else										return(ReadFrame(ih));
}

/* Close the file */
CGLOBAL int ID3v2CloseHandle(LPID3V2Handle ih)
{
	int err=ID3V2_ERR_OK;

	if (ih==NULL) return(err);

	// when writing, do some additional cleanup
	if (ih->Mode & ID3V2_MODE_WRITE)
		err = CloseWrite(ih);

	if (ih->oldfh) fclose(ih->oldfh);
	if (ih->fh)		fclose(ih->fh);
//	ih->fh = NULL;

	if (ih->FrameBuffer) free(ih->FrameBuffer);
//	ih->FrameBuffer = NULL;
	
	free(ih);
	return(err);
}

/* Remove an ID3v2 Tag from the file.
 * return: ID3V2_ERR_OK: the tag has been successfully removed
 *         ID3V2_ERR_NOID3V2: there isn't any ID3V2 Tag
 *         otherwise the usual error code
 */
CGLOBAL int ID3V2RemoveTag(LPCTSTR file)
{
	int err;
	FILE *fh=NULL;
	struct ID3V2_Header id3header;
	TCHAR tmpname[4096];

	// open the original file
	err = ID3V2_ERR_WRITE;
	if ((fh = _tfopen(file, _T("r+b"))) != NULL)
	{
		int headerLength;
		err = ID3V2_ERR_NOID3V2;
		if (ReadID3V2Header(fh, &id3header, &headerLength) == ID3V2_ERR_OK)
		{
			// get the buffer
			char *buffer = (char *)malloc(BUFSIZE);
			err = ID3V2_ERR_NOMEM;
			if (buffer != NULL)
			{
				FILE *targetFH;
				// seek to the position, where the actual data starts
				fseek(fh, headerLength, SEEK_CUR);
				// create a temporary filename, and then write to it
				_stprintf(tmpname, _T("%s.mp3tmp"), file);	// new filename
				// open the new file
				err = ID3V2_ERR_WRITE;
				if ((targetFH = _tfopen(tmpname, _T("wb"))) != NULL)
				{
					// now copy the data
					int readbytes = BUFSIZE;
					while (readbytes == BUFSIZE)
					{
						readbytes = fread(buffer, 1, BUFSIZE, fh);
						err = fwrite(buffer, 1, readbytes, targetFH);
						if (err != readbytes) return(ID3V2_ERR_WRITE);
					}
					fclose(targetFH);
					fclose(fh);
					fh = NULL;

					// delete the old file
					if (_tremove(file) != 0)
					{
						// if the first try failed, wait 0.3s, and try again, then error
						Sleep(300);
						if (_tremove(file) != 0)
						{
							return(ID3V2_ERR_DELETE);
						}
					}

					// move the new file to the place of the original file
					if (_trename(tmpname, file) != 0)
					{
						// if the first try failed, wait 0.3s, and try again, then error
						Sleep(300);
						if (_trename(tmpname, file) != 0)
						{
							return(ID3V2_ERR_MOVE);
						}
					}
				}
				free(buffer);
			}
		}
		if (fh) fclose(fh);
	}
	
	// everything is OK JLBOYER
	err = ID3V2_ERR_OK;
	return(err);
}

CGLOBAL int ID3v2CalcSize(int DataLength, int FrameCount)
{
	return(DataLength + FrameCount*sizeof(struct FrameHeader3));
}

/**************************************************************************
 * local functions
 **************************************************************************/

/* open Modes are
 * #define ID3V2_MODE_FAST 2		if set, read the complete Header in memory
 */
static int OpenRead (LPID3V2Handle ih)
{
	int err;
	struct ID3V2_Header id3header;

	err = ID3V2_ERR_FILE;
	if ((ih->fh = _tfopen(ih->FileName, _T("rb"))) == NULL) goto error;


	// I have read the header, now analyze it:
	if ((err = ReadID3V2Header(ih->fh, &id3header, &(ih->ID3v2Length))) != ID3V2_ERR_OK) goto error;

	ih->Offset = 0;	// keep track of the position in the file

	// check version
	err = ID3V2_ERR_NOID3V2;
	if (id3header.Version > ID3V2_VERSION) goto error;

	// The header is OK, so no get the info:
	ih->ID3v2Version = id3header.Version;
	ih->ID3v2Revision = id3header.Revision;
	ih->ID3v2Flags = id3header.Flags;

	// everything is OK
	err = ID3V2_ERR_OK;

error:
	return(err);
}

static int ReadFrame(LPID3V2Handle ih)
{
	int err, i;
	struct FrameHeader2 fh2;
	struct FrameHeader3 fh3;
	int len;


	err = ID3V2_ERR_NOFRAME;	// no more frames available
	if ((ih->ID3v2Length - ih->Offset) < (int)((ih->ID3v2Version == 2)?sizeof(fh2):sizeof(fh3))) goto error;

	// Read the Frame header
	err = ID3V2_ERR_READ;
	if (ih->ID3v2Version == 2)
	{	// ID3v2.2
		if (fread(&fh2, 1, sizeof(fh2), ih->fh) != sizeof(fh2))
			goto error;
		ih->Offset += sizeof(fh2);
		// convert V2 header to v3 header...
		memset(&fh3, 0, sizeof(fh3));
		// copy the ident
		memcpy(fh3.Ident, fh2.Ident, sizeof(fh2.Ident));
		// copy the size
	} else {	// ID3v2.3
		if (fread(&fh3, 1, sizeof(fh3), ih->fh) != sizeof(fh3))
			goto error;
		ih->Offset += sizeof(fh3);
	}

	// check, if it is a valid frame: Ident consists of chars [0-9A-Z]
	err = ID3V2_ERR_FRAME;
	for (i=0; i<TAGSIZE; i++)
	{
		if (isdigit(fh3.Ident[i]) ||
			 (isalpha(fh3.Ident[i]) && isupper(fh3.Ident[i])))
			continue;
		// no frame Identifier
		goto error;
	}

	// create the Length
	if (ih->ID3v2Version == 2)
	{	// ID3v2.2
		len  = fh2.Size[0]<<16;
		len += fh2.Size[1]<<8;
		len += fh2.Size[2];
	} else {	// ID3v2.3
		len  = fh3.Size[0]<<24;
		len += fh3.Size[1]<<16;
		len += fh3.Size[2]<<8;
		len += fh3.Size[3];
	}

	if (ih->FrameBufLen < len)	// allocate a bigger one
	{
		err = ID3V2_ERR_NOMEM;
		ih->FrameBufLen = 0;
		if (ih->FrameBuffer) free(ih->FrameBuffer);

		ih->FrameBuffer = (LPBUF)malloc(len);
		if (ih == NULL) goto error;	// no mem
		ih->FrameBufLen = len;
	}


	// now read the Buffer
	// check if there is enough space in the file
	err = ID3V2_ERR_CORRUPT;	// no more frames available
	if ((ih->ID3v2Length - ih->Offset) < len) goto error;

	err = ID3V2_ERR_READ;
	if (fread(ih->FrameBuffer, 1, len, ih->fh) != (size_t)len) goto error;

	ih->Offset += len;

	ih->FrameLength = len;

	// copy the header

	// convert the ID to the V3 format
	if (ih->ID3v2Version == 2)
	{
		int fid;
		for (fid=0; fid<ID3TAG_NUMTAGS; fid++)
			if (strncmp(fh3.Ident, ID3Frames[fid].ID2Str, ID3V22_FRAMEIDLEN) == 0)
			{
				strncpy(ih->FrameID, ID3Frames[fid].IDStr, ID3V2_FRAMEIDLEN);
				break;
			}
		if (fid >= ID3TAG_NUMTAGS) // not found
			strncpy(ih->FrameID, fh3.Ident, ID3V22_FRAMEIDLEN);
	}
	else
		strncpy(ih->FrameID, fh3.Ident, ID3V2_FRAMEIDLEN);


	// handle the flags fields
	ih->FrameFlags[0] = ih->FrameFlags[1] = 0;
	if (ih->ID3v2Version >= 3)
	{
		ih->FrameFlags[0] = fh3.flags[0];
		ih->FrameFlags[1] = fh3.flags[1];
	}

	err = ID3V2_ERR_OK;

	err = decodeFrame(ih);


error:
	return(err);
}

/** Unsynchronize the current frame, if necessary. */
static void unsynchronizeFrame(LPID3V2Handle ih)
{
	if (ih->ID3v2Flags & ID3V2H_FLAG_SYNC) {
		int flen = ih->FrameLength;
		int tpos = 0;	// target position, i am currently writing to
		LPBUF buf = ih->FrameBuffer;
		int idx = 0;
		for (idx=0; idx<(flen-1); idx++) {
			unsigned char c = buf[idx];
			unsigned char c2 = buf[idx+1];
			// check for compination 0xff00, which should be unsynchronized
			if (c == 0xff && c2 == 0x00) {
				idx++;	// skip the zero byte
			}
			// write the current byte
			buf[tpos++] = c;
		}
		// now we have to handle the last byte(s)
		// either the last byte was synchronized, then nothing has to be done
		// or the last byte has to be copied
		if (idx == (flen-1)) {
			buf[tpos++] = buf[idx];
		}
		// now set the actual frame-length
		ih->FrameLength = tpos;
	}
}

int uncompressFrame(LPBUF *buffer, int *buflen, int datalen);

static int decodeFrame(LPID3V2Handle ih)
{
	if (ih->FrameFlags[1] & ID3V2FH_FLAG2_COMPRESS)
	{
		ih->FrameLength = uncompressFrame(&ih->FrameBuffer, &ih->FrameBufLen, ih->FrameLength);
		if (ih->FrameLength < 0) return(ID3V2_ERR_CORRUPT);
	}
	return(ID3V2_ERR_OK);
}

/*
 * buffer = malloc() buffer
 * buflen = max size of buffer
 * return, the size of the uncompressed data, or -1 if failed
 */
int uncompressFrame(LPBUF *buffer, int *maxlen, int datalen)
{
	// frame requires uncompression
	// copy the data onto the stack
	int datalength = datalen-4;
	int newlength, error;
	char *data = (char *)alloca(datalength);
	if (data == NULL) return(-1);
	memcpy(data, (*buffer)+4, datalength);
	// check the new length of this frame
	newlength  = ((unsigned char *)(*buffer))[0]<<24;
	newlength += ((unsigned char *)(*buffer))[1]<<16;
	newlength += ((unsigned char *)(*buffer))[2]<<8;
	newlength += ((unsigned char *)(*buffer))[3];
	if (*maxlen < newlength)	// allocate a bigger one
	{
		*buffer = (LPBUF)realloc(*buffer, newlength);
		if (*buffer == NULL) return(-1);	// no mem
		*maxlen = newlength;
	}
	// now decode the data
	error = uncompress(*buffer, &newlength, data, datalength);
	if (error != Z_OK || newlength != datalen) return(-1);
	return(newlength);
}

static int compressFrame(LPID3V2Handle ih)
{
	int newlength, error, datalength;
	char *data;
	// check if we should compress the frame
	if ((ih->MinCompressLength == -1) || (ih->FrameLength < ih->MinCompressLength))
		return(ID3V2_ERR_OK);
	// OK, then I'll compress the frame....
	// copy the data onto the stack
	datalength = ih->FrameLength;
	data = (char *)alloca(datalength);
	if (data == NULL) return(ID3V2_ERR_OK);	// no comppression....
	memcpy(data, ih->FrameBuffer, datalength);
	// check the new length of this frame
	newlength  = datalength + datalength/10 + 12 + 4;	// additional 4 bytes for the uncompressed length
	if (ih->FrameBufLen < newlength)	// allocate a bigger one
	{
		ih->FrameBuffer = (LPBUF)realloc(ih->FrameBuffer, newlength);
		if (ih->FrameBuffer == NULL) return(ID3V2_ERR_NOMEM);	// no mem
		ih->FrameBufLen = newlength;
	}
	// now decode the data
	error = compress(ih->FrameBuffer+4, &newlength, data, datalength);
	if (error != Z_OK || datalength <= (newlength+4))
	{
		// reset to the default
		memcpy(ih->FrameBuffer, data, datalength);
	}
	else
	{
		ih->FrameBuffer[0] = ((datalength>>24) & 0xFF);
		ih->FrameBuffer[1] = ((datalength>>16) & 0xFF);
		ih->FrameBuffer[2] = ((datalength>>8 ) & 0xFF);
		ih->FrameBuffer[3] = ((datalength)     & 0xFF);
		ih->FrameLength = newlength + 4;
		ih->FrameFlags[1] |= ID3V2FH_FLAG2_COMPRESS;
	}
	return(ID3V2_ERR_OK);
}


/*
#define ID3V2_MODE_HAVELEN 2		/* length of Tag is already known
#define ID3V2_MODE_COMPACT 4		/* do not add additional space in the header
*/
static int OpenWrite(LPID3V2Handle ih)
{
	int err;
	FILE *fh=NULL;
	struct ID3V2_Header id3header;
	short Overwrite=FALSE;	// flag, if I can overwrite the tag in the old file


	// open the original file
	err = ID3V2_ERR_WRITE;
	if ((fh = _tfopen(ih->FileName, _T("r+b"))) == NULL) goto error;

	if (ReadID3V2Header(fh, &id3header, &(ih->OldID3V2Length)) == ID3V2_ERR_OK)
	{
		// A previous ID3V2 header is found. Check if I can overwrite it
		if ((ih->Mode & ID3V2_MODE_HAVELEN) && 
			 (ih->OldID3V2Length >= ih->ID3v2Length))
		{
			// I have enough space to overwrite the old Tag
			Overwrite = TRUE;
			fseek(fh, 0, SEEK_SET);	// seek at the beginning, so I can start
											// writing
		}
		else
		{
			// I have an ID3v2 Tag, but it is too small
			// so seek at the position, where the actual Data starts
			fseek(fh, ih->OldID3V2Length, SEEK_CUR);
		}
	}
	else
	{
		// no ID3V2; so seek back to the beginning
		fseek(fh, 0, SEEK_SET);
	}

// Set the version in the structure
	// write current version in the header
	strncpy(id3header.Ident, IDENTSTRING, ID3V2_FRAMEIDLEN);
	id3header.Version = ih->ID3v2Version = ID3V2_VERSION;
	id3header.Revision = ih->ID3v2Revision = ID3V2_REVISION;
	id3header.Flags = 0;
	ih->ID3v2Length = 0;				// count here the size of the Tag

	// the header is now built
	// if Overwrite, then the length is OK too, else I have to set it when
	// I close the file.

	if (!Overwrite)
	{
		// I have to create a new file, and delete the old one...
		ih->oldfh = fh;

		// create a temporary filename, and then write to it

		err = ID3V2_ERR_NOMEM;
		_stprintf((LPTSTR)ih->TempName, _T("%s.mp3tmp"), ih->FileName);	// new filename

		// open the new file
		err = ID3V2_ERR_WRITE;
		if ((fh = _tfopen(ih->TempName, _T("wb"))) == NULL) goto error;
		id3header.Size[0] = id3header.Size[1] = id3header.Size[2] = id3header.Size[3] = 0;
	}

	// write the header
	fwrite(&id3header, 1, sizeof(id3header), fh);

// fh is now the filehandle to write to
	ih->fh = fh;
	fh = NULL;

	err = ID3V2_ERR_OK;

error:
	if (fh) fclose(fh);
	return(err);
}

char *_ID3v2PaddingString = "MP3ext ";

static int CloseWrite(LPID3V2Handle ih)
{
	int i, err;
	int readbytes;
	char *buffer = (char *)malloc(BUFSIZE);
	char *paddingString = ih->PaddingString;
	if (paddingString == NULL) paddingString = ID3V2_PADDING_STRING;
	if (paddingString == NULL) paddingString = "EZMP3  ";

	if (buffer == NULL)
		return(ID3V2_ERR_NOMEM);

	/* check, wether we are writing in a new file, or if
	 * we simply replace the old ID3v2-Header */
	if (ih->oldfh && ih->fh)	// check, if everything is fine
	{
		/* This is a new file, so now, we have to do
		 * 1. add some space for a buffer (size is in the global Var ID3v2PaddingSize)
		 * 2. copy the data from the old handle to the new file; 
		 *    oldfh points to this data
		 * 3. Adjust the size in the ID3v2-Header
		 * 4. Close the new file.
		 * 5. delete the old file
		 * 6. move the new file to the place of the original file
		 * finished
		 */
	
		// 1. add some space for a buffer
		if (!(ih->Mode & ID3V2_MODE_COMPACT))
		{
			// write ID3v2PaddingSize zero-bytes in the file
			for (i=0; i<ih->PaddingSize; i++)
			{
				char c;
				if (i==0) 
					c = '\0';
				else
					c = '\0'; //paddingString[(i-1)%strlen(paddingString)];

				err = fputc(c, ih->fh);
				if (err == EOF) {
					err = ID3V2_ERR_WRITE;
					goto error;
				}
				ih->ID3v2Length++;	// increase the Length of the Tag
			}
		}

		// 2. copy the data from the old handle to the new file; 
		//    oldfh points to this data

		assert(ih->oldfh);
		readbytes = BUFSIZE;
		while (readbytes == BUFSIZE)
		{
			readbytes = fread(buffer, 1, BUFSIZE, ih->oldfh);
			err = fwrite(buffer, 1, readbytes, ih->fh);
			if (err != readbytes) {
				err = ID3V2_ERR_WRITE;
				goto error;
			}
		}
		// check if an error occured
		if (!(feof(ih->oldfh))) {
			err = ID3V2_ERR_READ;
			goto error;
		}


		// 3. Adjust the size in the ID3v2-Header
		{
			struct ID3V2_Header id3header;
			//	fill in the Header
			strncpy(id3header.Ident, IDENTSTRING, ID3V2_FRAMEIDLEN);
			id3header.Version = ih->ID3v2Version;
			id3header.Revision = ih->ID3v2Revision;
			id3header.Flags = 0;
			id3header.Size[0] = ((ih->ID3v2Length>>21) & 0x7F);
			id3header.Size[1] = ((ih->ID3v2Length>>14) & 0x7F);
			id3header.Size[2] = ((ih->ID3v2Length>>7)  & 0x7F);
			id3header.Size[3] = ((ih->ID3v2Length)     & 0x7F);
			fseek(ih->fh, 0, SEEK_SET);
			fwrite(&id3header, 1, sizeof(id3header), ih->fh);
		}
		// 4. Close the files.
		fclose(ih->fh);
		ih->fh = NULL;
		fclose(ih->oldfh);
		ih->oldfh = NULL;

		// 5. delete the old file
		if (_tremove(ih->FileName) != 0)
		{
			// if the first try failed, wait 0.3s, and try again, then error
			Sleep(300);
			if (_tremove(ih->FileName) != 0)
			{
				err = ID3V2_ERR_DELETE;
				goto error;
			}
		}

		// 6. move the new file to the place of the original file
		if (_trename(ih->TempName, ih->FileName) != 0)
		{
			// if the first try failed, wait 0.3s, and try again, then error
			Sleep(400);
			if (_trename(ih->TempName, ih->FileName) != 0)
			{
				err = ID3V2_ERR_MOVE;
				goto error;
			}
		}
	}
	else
	{
		// we are overwriting the old tag, so now clean of the
		// previous data
		for (i=ih->ID3v2Length; i<ih->OldID3V2Length; i++)
		{
			char c = '\0'; //paddingString[(i-ih->ID3v2Length)%strlen(paddingString)];
			fputc(c, ih->fh);
		}
	}
	err = ID3V2_ERR_OK;

error:
	if (buffer) free(buffer);
	return(err);
}

/***********************************************************+
 * WriteFrame
 *
 * Write the specific Frame into the file. The following fields
 * of ih are affected:
 *
 * fh:          (w) File which is written into
 * ID3v2Length: (w) increased by the size of the Frame incl header
 * FrameID:     (r) ID of the frame
 * FrameLength: (r) length of the frame
 * FrameBuffer: (r) the actual data to write
 * OldID3V2Length: max length, if I am in overwrite-mode (oldfh==NULL)
 */
static int WriteFrame(LPID3V2Handle ih)
{
	int err;
	struct FrameHeader3 fheader;

	memset(&fheader, 0, sizeof(fheader));

	compressFrame(ih);	// do the compression, if necessary

	// if in overwrite-mode, chek if there is enough space
	err = ID3V2_ERR_OLDSIZE;
	if ((ih->oldfh == NULL) &&
		 ((ih->ID3v2Length + ih->FrameLength + (int)sizeof(fheader)) > ih->OldID3V2Length))
		goto error;

	

	// now create the frame header
	strncpy(fheader.Ident, ih->FrameID, ID3V2_FRAMEIDLEN);
	// create the Length
	fheader.Size[0] = (((ih->FrameLength)>>24) & 0xFF);
	fheader.Size[1] = (((ih->FrameLength)>>16) & 0xFF);
	fheader.Size[2] = (((ih->FrameLength)>>8)  & 0xFF);
	fheader.Size[3] =  ((ih->FrameLength)      & 0xFF);
	// the flags:
	fheader.flags[0] = ih->FrameFlags[0];
	fheader.flags[1] = ih->FrameFlags[1];

	// write the frame-header
	err = ID3V2_ERR_WRITE;
	if (fwrite(&fheader, 1, sizeof(fheader), ih->fh) != sizeof(fheader))
		goto error;
	ih->ID3v2Length += sizeof(fheader);


	// write the data
//	err = ID3V2_ERR_WRITE;
	if (fwrite(ih->FrameBuffer, 1, ih->FrameLength, ih->fh) != (size_t)ih->FrameLength)
		goto error;
	ih->ID3v2Length += ih->FrameLength;

	err = ID3V2_ERR_OK;

error:
	return(err);
}

/* Read the Tag from the file, an check if the header is OK
 */
static int ReadID3V2Header(FILE *fh, struct ID3V2_Header *id3header, int *ID3V2size)
{
	int err;

	assert(fh);

	*ID3V2size = 0;	// this is for the size

	err = ID3V2_ERR_READ;
	if (fread(id3header, 1, sizeof(struct ID3V2_Header), fh) != sizeof(struct ID3V2_Header))
		goto error;

	// I have read the header, now analyze it:
	err = ID3V2_ERR_NOID3V2;

	if ((strncmp(id3header->Ident, IDENTSTRING, 3) != 0) ||
		 (id3header->Size[0] & 0x80) ||
		 (id3header->Size[1] & 0x80) ||
		 (id3header->Size[2] & 0x80) ||
		 (id3header->Size[3] & 0x80))
		goto error;


	err = ID3V2_ERR_OK;

	// now the size:
	if (ID3V2size != NULL) {
		*ID3V2size  = (id3header->Size[0]<<21) +
						  (id3header->Size[1]<<14) +
						  (id3header->Size[2]<<7) +
						  (id3header->Size[3]);
	}

error:
	return(err);

}
