/*****************************************************************************
    This file is part of MP3ext.
    Copyright (C) 1999  Michael Mutschler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    email: michael@mutschler.de
******************************************************************************/
/*
 * Header-file for the ID3V2 functions 
 */

#ifndef _ID3V2_H
#define _ID3V2_H

#include <stdio.h>
#include "ID3V2Errors.h"

/* The following version of ID3V2 is supported by these functions: */
#define ID3V2_VERSION 3
#define ID3V2_REVISION 0

/* For the String-fields: The type of encoding */
#define ID3V2_ANSI 0
#define ID3V2_UNICODE 1

/* this defines the length of a frame-identifier */
#define ID3V22_FRAMEIDLEN 3
#define ID3V2_FRAMEIDLEN 4

/* size of the ID3v2 header */
#define ID3V2_HEADER_LENGTH 10

/* type of the frame contents */

#ifndef LPBUF
typedef unsigned char * LPBUF;
#endif
#ifndef MAX_FILE
#define MAX_FILE 1024
#endif

/* flags defined for the ID3v2 header. */
#define ID3V2H_FLAG_SYNC (1<<7)
#define ID3V2H_FLAG_COMPRESSION (1<<6)

/* structure for scanning the ID3V2 header. Do not allocate it on your own!!! */
/* (r) means only available during reading, else available during reading and writing */
typedef struct _ID3V2Handle {
/* global part */
	int   ID3v2Length;			/* this is a read-only variable */
	char  ID3v2Version;			/* Version of the Tag (r) */
	char  ID3v2Revision;		/* Revision of the Tag (r) */
	unsigned char  ID3v2Flags;	/* a copy of the flags-field of the header. */
/* variables specific to the frames */
	char  FrameID[ID3V2_FRAMEIDLEN];	/* Identifier of this frame */
	unsigned char	FrameFlags[2];		/* the two flag bytes of the current frame */
	int   FrameLength;			/* length of this frame */
	LPBUF FrameBuffer;			/* the actual content of the frame
										 * the buffer is provided by the ID3v2 functions */
	int   PaddingSize;			/* after opening, you can set this Var to the padding size */
	int	MinCompressLength;	/* minimum size for frame to be compressed.
										 * -1: none are compressed (default)
										 *  0: all are compressed
										 */
	char *PaddingString;			/* fill the padding area with repetitions of this string. */
/* private stuff */
   int   State;         /* state where I currently am with writing */
	int	Mode;          /* store the access-mode here */
	FILE *fh;				/* I have to store the filepointer somewhere */
	int	Offset;			/* keep track of where I am in the file */
	int	FrameBufLen;   /* Length of the currently allocated FrameBuffer */
	const TCHAR FileName[MAX_FILE];	/* original filename */
	const TCHAR TempName[MAX_FILE];	/* temporary filename */
	int   OldID3V2Length;/* This gets the length of the previous length of the
								 * ID3V2 Header during writing */
	FILE *oldfh;         /* filehandle of the original file; if this is not NULL,
	                      * then I am in copy-mode. It points to the actual MP3-Data */
} ID3V2Handle, *LPID3V2Handle;

// frameheader flags
#define ID3V2FH_FLAG1_TAG_PRESERVE (1<<7)
#define ID3V2FH_FLAG1_FRAME_PRESERVE (1<<6)
#define ID3V2FH_FLAG1_READONLY (1<<5)
#define ID3V2FH_FLAG2_COMPRESS (1<<7)	// this one is used
#define ID3V2FH_FLAG2_ENCRYPT (1<<6)
#define ID3V2FH_FLAG2_GROUP (1<<5)

/***************************************************************************/
/* Functions
/***************************************************************************/

#ifndef CGLOBAL
#ifdef __cplusplus
#define CGLOBAL extern "C"
#else
#define CGLOBAL extern
#endif
#endif

/* This is the global function to read/write the MP3Info */
CGLOBAL LPID3V2Handle ID3V2OpenFile(LPCTSTR file, int mode, int *ErrorCode, int data);
/* process the next frame */
CGLOBAL int ID3v2NextFrame(LPID3V2Handle ih);
/* Close the file */
CGLOBAL int ID3v2CloseHandle(LPID3V2Handle ih);
/* Function to calculate the size of a Tag, when given the total length
 * of all frames, and the number of frames */
CGLOBAL int ID3v2CalcSize(int DataLength, int FrameCount);
/* Remove an ID3v2 Tag from the file */
CGLOBAL int ID3V2RemoveTag(LPCTSTR file);

/* These are the internal states I have for the reading/writing */
enum {
	ST_INIT,
	ST_WRITING,
	ST_READING,
};

/* The string which is used for padding */
CGLOBAL char *_ID3v2PaddingString;
#define ID3V2_PADDING_STRING _ID3v2PaddingString


/* modes for Opening */
#define ID3V2_MODE_READ 0			/* bit 0 - read/write */
#define ID3V2_MODE_WRITE 1
/*  flags specific to the write-process */
#define ID3V2_MODE_HAVELEN 2		/* length of Tag is already known
                                  * Use ID3v2CalcSize() to calculate it, and pass it
											 * in data when opening */
#define ID3V2_MODE_COMPACT 4		/* do not add additional space in the header */
/* flags specific to the read-process */
#define ID3V2_MODE_FAST 2			/* if set, read the complete Header in memory */

#endif /* _ID3V2_H */
