/*****************************************************************************
    This file is part of MP3ext.
    Copyright (C) 1999  Michael Mutschler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    email: michael@mutschler.de
******************************************************************************/
/*
 * MP3info.cpp
 */

#include "priv.h"
#include "definition.h"		// header EZ3
#include "MP3Info.h"
#include "MP3PicFrame.h"
#include "StringConverter.h"
#include <sys/stat.h>		// use stat for file-info


extern bool IsStringEmpty(const string& s);
extern bool IsStringEmpty(string *s);
extern bool IsStringEmpty(wstring *s);
extern bool IsStringEmpty(const wstring& s);



#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#endif

/* ******************************************************************** */
CMP3Info::CMP3Info(tstring& file, bool scanfile)
{
	
	//
	// Information from EZ3 playlist format file
	//
	if (scanfile == false)
	{	
		// Read from playtlist EZ3
		// See InsertFileMP3 for more information
		m_Filename = file;		// set the FileName
		m_id3v1 = NULL;
		m_tagVersion = VERSION_ID3NO;
		return;
	}


	//
	// Scan file to get MP3 & core informations
	// 
	m_Filename = file;				// set the FileName
	m_id3v1 = NULL;					// not yet id3v1
	m_mp3Size = 0;					// not yet
	m_tagVersion = VERSION_ID3NO;	// no tag yet
	
	memset(&m_CoreMP3Info, 0, sizeof(m_CoreMP3Info));
	
	
	int id3v2ret, nID3v2Length = 0;
	
	struct _stat sinfo;
	int err = 0;

	err = _tstat(file.c_str(), &sinfo);
	if (err == 0)
	{
		/* get the size of the file */
		m_CoreMP3Info.filesize = sinfo.st_size;
		m_mp3Size = m_CoreMP3Info.filesize;
	}



	DWORD attr = GetFileAttributes(file.c_str());



	if (attr != 0xFFFFFFFF)
	{

		m_ReadOnly = (attr & FILE_ATTRIBUTE_READONLY);

		// ///////////////////////////////
		// read ID3V2
		id3v2ret = ReadID3V2(&nID3v2Length);
		
		if (id3v2ret == ID3V2_ERR_OK)
		{
			m_tagVersion |= VERSION_ID3V2;
			m_mp3Size -= nID3v2Length;

		}
			
		// ///////////////////////////////
		// read ID3V1
		if (ReadID3V1() == ID3V2_ERR_OK) 
		{
			m_tagVersion |= VERSION_ID3V1;
			m_mp3Size -= 128;
		}
		
		// ///////////////////////////////
		// Read the CoreMP3Info for mpeg1,2,3
		mpg123_readinfo(GetCFilename(), &m_CoreMP3Info, HaveID3v1());
		
	}

	// If no tag found add the filename in the Artist frame
	ParseFilename();

	// handle the size 
	m_diffSize = 0;	// default. No diff
	// now, make sure, we have a TSIZ-frame
	CMP3StringFrame *mp3f = GetStringFrame(ID3TAG_TSIZ);
	if (mp3f != NULL)	
	{
		// I have a TSIZ-frame. so calculate the diff
		int tsiz = _ttoi(mp3f->GetString().c_str());
		m_diffSize = m_mp3Size - tsiz;
	}

	// other settings
	m_changed = FALSE;
}

/* ******************************************************************** */
CMP3Info::~CMP3Info()
{
	DeleteTag(VERSION_ID3V1 + VERSION_ID3V2);
}


/* ******************************************************************** */
void CMP3Info::ParseFilename()
{
	// If no tag found add the filename in the Artist frame
	if (m_tagVersion == 0)
	{
		TCHAR *p, szTemp[1024];

		p = _tcsrchr(GetCFilename(), '\\');

		if (p)
			_tcscpy(szTemp, p+1);
		else
			_tcscpy(szTemp, GetCFilename());

		p = szTemp;
		*p = toupper(*p);

		tstring str(szTemp);

		CMP3StringFrame *sf = new CMP3StringFrame(ID3TAG_TPE1, str);
		if (sf) 
			m_FrameArray.push_back(sf);
	
	}
}


/* ******************************************************************** */
int CMP3Info::ReadID3V1()
{
	struct ID3V1Info id3i;
	
	if (ReadID3V1Info(GetCFilename(), &id3i) != 0) 
		return(ID3V2_ERR_NOMEM);	// error reading ID3V1 Info

	// I have an initialized ID3V1Info-Structure
	// now translate it in the internal form
	if (m_id3v1 != NULL)
		free(m_id3v1);


	m_id3v1 =  (struct ID3V1MP3Info*)calloc(sizeof(struct ID3V1MP3Info), 1);
	if (!m_id3v1)
		goto error;

	// Title Song name
	memcpy(m_id3v1->SongName, &id3i.SongName,sizeof(id3i.SongName));
	m_id3v1->SongName[30] = '\0';
	
	// Artist
	memcpy(m_id3v1->Artist, &id3i.Artist, sizeof(id3i.Artist));
	m_id3v1->Artist[30] = '\0';
	
	// Album
	memcpy(m_id3v1->Album, &id3i.Album, sizeof(id3i.Album));
	m_id3v1->Album[30] = '\0';
	
	// Year
	memcpy(m_id3v1->Year, &id3i.Year, sizeof(id3i.Year));
	m_id3v1->Year[4] = '\0';
	
	// Comment
	memcpy(m_id3v1->Comment, &id3i.Comment, sizeof(id3i.Comment));
	m_id3v1->Year[4] = '\0';

	// Genre
	m_id3v1->Genre = id3i.Genre;
	
	// Track only for ID3v1.1
	m_id3v1->Track = id3i.Track;

	return(ID3V2_ERR_OK);

error:
	return(ID3V2_ERR_NOMEM);
}

/* ******************************************************************** */
int CMP3Info::ReadID3V2(int *nID3v2Length)
{
	int err, fid;
	LPID3V2Handle ih;

	CMP3Frame *mp3f = NULL;

	if (!(ih = ID3V2OpenFile(GetCFilename(), ID3V2_MODE_READ, &err, 0)))
		goto error;

	*nID3v2Length = ih->ID3v2Length + ID3V2_HEADER_LENGTH;

	while((err=ID3v2NextFrame(ih)) == ID3V2_ERR_OK)
	{
		for (fid=0; fid<ID3TAG_NUMTAGS; fid++)
			if (strncmp(ih->FrameID, ID3Frames[fid].IDStr, ID3V2_FRAMEIDLEN) == 0) break;

		if (fid == ID3TAG_NUMTAGS ||	// unknown TAG
			 !ID3Frames[fid].Supported	// unsupported
			) continue;	// just skip it

		// now I have a supported TAG. So add it
		err = ID3V2_ERR_NOMEM;
		if (!(mp3f = CreateMP3Frame(fid))) goto error;

		err = mp3f->SetID3v2Buffer(ih->FrameBuffer, ih->FrameLength, ih->ID3v2Version);
		if (err != ID3V2_ERR_OK) goto error;

		AddFrame(mp3f);
		mp3f = NULL;
	}

	
	err = ID3V2_ERR_OK;

error:
	if (mp3f) delete mp3f;
	if (ih) ID3v2CloseHandle(ih);
	return(err);
}


/* ******************************************************************** */
// get Tag after start-Tag
int CMP3Info::FindFrame(int ID, int subID, int start)
{
	int fsize = m_FrameArray.size();
	for (int i=(start+1); i<fsize; i++)
	{
		if (m_FrameArray[i]->GetID() == ID &&
			 (subID == -1 ||
			  m_FrameArray[i]->getSubID() == subID
			 )
			)
			return(i);
	}
	// not found
	return(-1);
}

/* ******************************************************************** */
int CMP3Info::FindFrame(CMP3Frame *mp3f, int start)
{
	int fsize = m_FrameArray.size();
	for (int i=(start+1); i<fsize; i++)
	{
		if (m_FrameArray[i] == mp3f)
			return(i);
	}
	// not found
	return(-1);
}

/* ******************************************************************** */
// just to get a tstring-frame
CMP3StringFrame *CMP3Info::GetStringFrame(int ID, int start)
{
	assert(ID3Frames[ID].Type == ID3ENC_STRING);
	// check if the ID is OK
	if (ID3Frames[ID].Type != ID3ENC_STRING) return(NULL);
	// now get the actual frame
	return((CMP3StringFrame *)GetFrame(ID, -1, start));
}

/* ******************************************************************** */
// Apply all changes to the file
int CMP3Info::SaveTag()
{
	int err = 0;
	
	err = WriteID3V2();

	if (err == ID3V2_ERR_NOID3V2)	err = ID3V2_ERR_OK;

	if (err != ID3V2_ERR_OK)
		return(err);

	err = WriteID3V1();

	// set the flag to unchanged
	m_changed = FALSE;

	return(err);
}

/* ******************************************************************** */
int CMP3Info::WriteID3V1()
{
	// If there is no Tag, then do a remove
	if (!HaveID3v1())
	{
		return(WriteID3V1Info(GetCFilename(), NULL, ID3V1_WRITE_REMOVE));
	}

	struct ID3V1Info id3i;
	memset (&id3i, 0, sizeof(struct ID3V1Info));


	// Title Song name
	memcpy(&id3i.SongName, m_id3v1->SongName, sizeof(id3i.SongName));
	
	// Artist
	memcpy(&id3i.Artist, m_id3v1->Artist, sizeof(id3i.Artist));
	
	// Album
	memcpy(&id3i.Album, m_id3v1->Album, sizeof(id3i.Album));
	
	// Year
	memcpy(&id3i.Year, m_id3v1->Year, sizeof(id3i.Year));
	
	// Comment
	memcpy(&id3i.Comment, m_id3v1->Comment, sizeof(id3i.Comment));

	// Genre
	id3i.Genre = m_id3v1->Genre;
	
	// Track only for ID3v1.1
	id3i.Track = m_id3v1->Track;


	return(WriteID3V1Info(GetCFilename(), &id3i, ID3V1_WRITE_NORMAL));

}

/* ******************************************************************** */
int CMP3Info::WriteID3V2()
{
	int fnr, tsize, err;
	LPID3V2Handle ih;
	CMP3Frame *mp3f;

	// If there is no Tag, then do a remove
	if (!HaveID3v2())
	{
		return(ID3V2RemoveTag(GetCFilename()));
	}

	// before saving, check if a TSIZ Tag exists; if not, create a new one:
	CMP3StringFrame *mp3sf = GetStringFrame(ID3TAG_TSIZ);
	if (mp3sf == NULL)		// create a new one, if it doesn't exist
	{
		mp3sf = (CMP3StringFrame *)CreateMP3Frame(ID3TAG_TSIZ);
		if (mp3sf != NULL)
		{
			// set the size as an ascii-tstring
			TCHAR buf[20];
			_itot(m_mp3Size, buf, 10);
			mp3sf->SetString(buf);
			ReplaceFrame(mp3sf);
		}
	}

	/* currently I have to know, how big the Tag is going to be
	 */
	tsize = 0;
	for (fnr = 0; fnr < m_FrameArray.size(); fnr++)
	{
		mp3f = m_FrameArray[fnr];
		tsize += mp3f->GetID3v2Length();
	}
	// Calc the actual size:
	tsize = ID3v2CalcSize(tsize, m_FrameArray.size());
	if (ih = ID3V2OpenFile(GetCFilename(), ID3V2_MODE_WRITE|ID3V2_MODE_HAVELEN,  &err, tsize))
	{
		ih->PaddingSize = 128;	// set padding to 32 bytes
		ih->MinCompressLength = -1;	// no compression
		// write all the tags
		for (fnr=0; fnr<m_FrameArray.size(); fnr++)
		{
			mp3f = m_FrameArray[fnr];
			// fill in the FrameData in teh structure
			strncpy(ih->FrameID, ID3Frames[mp3f->GetID()].IDStr, ID3V2_FRAMEIDLEN);
			ih->FrameLength = tsize = mp3f->GetID3v2Length();

			// now the buffer
			if (tsize > ih->FrameBufLen)
			{	// allocate a new buffer
				if (ih->FrameBuffer != NULL) free(ih->FrameBuffer);
				ih->FrameBufLen = tsize + 1024;	// always use a kb more
				ih->FrameBuffer = (LPBUF)malloc(ih->FrameBufLen);
			}
			if (ih->FrameBuffer == NULL) break;	// faild to increase the buffer

			err = mp3f->GetID3v2Buffer(ih->FrameBuffer, ih->FrameBufLen);
			if (err != ID3V2_ERR_OK) break;

			err = ID3v2NextFrame(ih);
			if (err != ID3V2_ERR_OK) break;
		}
		// finished. Close everything

		if (err == ID3V2_ERR_OK)
			err = ID3v2CloseHandle(ih);		
		else	
			ID3v2CloseHandle(ih);	// keep previous error-code
	}

	return(err);
}

/* ******************************************************************** */
// Add a frame. if replace==TRUE, first one is replaced
bool CMP3Info::AddFrame(CMP3Frame *mp3f, bool replace)
{
	assert(mp3f);
	if (mp3f == NULL) return(FALSE);

	int index=-1;
	if (replace)
	{
		index = FindFrame(mp3f->GetID());
	}
	// now, if I don't have an index for replacement, add one
	if (index == -1)
		m_FrameArray.push_back(mp3f);
	else
	{
		delete m_FrameArray[index];	// remove old Frame
		m_FrameArray[index] = mp3f;	// replace old
	}
	// success
	m_changed = TRUE;

	return(TRUE);
}

/* ******************************************************************** */
/* Update a frame
 *
 * is new frame default? -> yes: if old exists -> yes: remove old
 *                                                no: return(FALSE)
 *                       -> no:
 * if old frame exists -> no: add a clone
 *                     -> yes:
 * contents are equal -> no: replace info
 *                    -> yes: return FALSE
 */
bool CMP3Info::UpdateFrame(CMP3Frame *mp3f)
{
	assert(mp3f);
	if (mp3f == NULL) return(FALSE);

// 1. find an old frame of same type
	int index=-1;
	CMP3Frame *omp3f;
	omp3f = GetFrame(mp3f->GetID(), mp3f->getSubID());	// get a replaceable frame

// check if the new frame is the default
	if (mp3f->IsDefault())
	{
		if (omp3f)
			return(DeleteFrame(mp3f->GetID(), omp3f));
		else
			return(false);
	}

// we have some useful content in the new frame
	if (omp3f == NULL)	// no previous found
	{
		omp3f = mp3f->Clone();
		if (omp3f) {
			if (AddFrame(omp3f))
				return(TRUE);	// something has changed
			else
				delete omp3f;
		}
		return(FALSE);		// error
	}

	// omp3f contans the old MP3Frame, which is to be replaced

// 2. check if they are equal
	if (*omp3f == *mp3f)
		return(FALSE);		// they are, so no change

// finally, copy the contents of the new frame to the old frame

	*omp3f = *mp3f;

	m_changed = TRUE;
	return(TRUE);
}

/* ******************************************************************** */
// delete a frame, if mp3f==NULL, then the first one found
bool CMP3Info::DeleteFrame(int ID, CMP3Frame *mp3f)
{
	int index = -1;
	if (mp3f == NULL)
		index = FindFrame(ID);
	else
		index = FindFrame(mp3f);

	if (index == -1) return(FALSE);		// error, not found

	// remove old frame
	delete m_FrameArray[index];
	m_FrameArray.erase(m_FrameArray.begin() + index);

	m_changed=TRUE;		// something has changed
	if (m_FrameArray.size() == 0)
		m_tagVersion &= ~VERSION_ID3V2;	

	return(TRUE);		// successful removed
}

/* ******************************************************************** */
bool CMP3Info::DeleteTag(int tagDel)
{
	
	
	if ((tagDel & VERSION_ID3V2) == VERSION_ID3V2)
	{
		for (int i=0; i<m_FrameArray.size(); i++)
			delete m_FrameArray[i];
		m_FrameArray.erase(m_FrameArray.begin(), m_FrameArray.end());

		m_changed = TRUE;
		m_tagVersion &= ~VERSION_ID3V2;	// Retranche ID3v2
	}

	if ((tagDel & VERSION_ID3V1) == VERSION_ID3V1)
	{
		if (m_id3v1 != NULL)
			free(m_id3v1);
		
		m_id3v1 = NULL;
		m_changed = TRUE;
		m_tagVersion &= ~VERSION_ID3V1;
	}
	
	return(TRUE);
}

/* ******************************************************************** */
void CMP3Info::sortFrames()	// sort the frames
{
	int size = m_FrameArray.size();
	for (int i=0; i<(size-1); i++)
	{
		for (int j=i; j<size; j++)
		{
			CMP3Frame *curr = m_FrameArray[j];
			CMP3Frame *fmin = m_FrameArray[i];
			if (ID3Frames[curr->GetID()].priority < ID3Frames[fmin->GetID()].priority)
			{
				m_FrameArray[j] = fmin;
				m_FrameArray[i] = curr;
			}
		}
	}
}


/* ******************************************************************** */
void CMP3Info::SetFileEZ3(EZ3HEADERMP3INFO *headerfile)	
{
	m_changed = FALSE;
	m_ReadOnly = FALSE;

	m_mp3Size = headerfile->mp3size;	
	m_diffSize = headerfile->diffsize;	
	m_tagVersion = headerfile->tagVersion;			
	m_CoreMP3Info = headerfile->coreMP3;
}

/* ******************************************************************** */
void CMP3Info::GetFileEZ3(EZ3HEADERMP3INFO *headerfile)	
{
	int nFrameStringOnly = 0;

	m_changed = FALSE;
	m_ReadOnly = FALSE;

	// Ne sauvegarde dans la plyalist EZ3 que les frame de type STRING
	// Compte uniquement les frames de type STRING uniquement
	for (int i=0; i< GetFrameCount(); i++)
	{
		CMP3Frame *mp3f = this->GetIndexFrame(i);
		if (mp3f)
		{
			if(mp3f->GetType() == ID3ENC_STRING)
				nFrameStringOnly++;
		}
	}

	headerfile->nCountFrame = nFrameStringOnly;	
	headerfile->tagVersion = m_tagVersion;			
	headerfile->mp3size = m_mp3Size ;	
	headerfile->diffsize = m_diffSize ;	
	headerfile->nLenFilename= _tcslen(GetCFilename());			
	headerfile->coreMP3 = m_CoreMP3Info;
}

/* ******************************************************************** */
BOOL CMP3Info::SetID3v1(struct ID3V1MP3Info *id3v1New)
{
	// Remove old ID3v1 if exist
	if(m_id3v1)
		free(m_id3v1);
	
	
	if (id3v1New == NULL)
	{	// Si pas de nouveau ID3v1 alors retire ID3v1, Pas d'ajout
		m_tagVersion &= ~VERSION_ID3V1;
		return TRUE;
	}

	// Create new ID3v1
	m_id3v1 =  (struct ID3V1MP3Info*)malloc(sizeof(struct ID3V1MP3Info));
	if (!m_id3v1)
		return (FALSE);		

	// Copy the new one
	memcpy(m_id3v1, id3v1New, sizeof(struct ID3V1MP3Info));


	m_tagVersion |= VERSION_ID3V1;

	return (TRUE);
}



/* ******************************************************************** */
void CMP3Info::CopyAll(CMP3Info *mp3src)
{

	DeleteTag(VERSION_ID3V1+VERSION_ID3V2);

	// Copy ID3v1
	struct ID3V1MP3Info *id3v1src = mp3src->GetID3v1();
	SetID3v1(id3v1src);
	
	// Copy ID3v2
	CMP3Frame *mp3f = NULL;

	int nCountFrame = mp3src->GetFrameCount();

	for (int i = 0; i<nCountFrame; i++)
	{
		CMP3Frame *nf; 
		mp3f = mp3src->GetIndexFrame(i);

		nf = mp3f->Clone();
		AddFrame(nf, FALSE);
		mp3f = NULL;
	}

	SetTagVersion(mp3src->GetTagVersion());
}

/* ******************************************************************** */
void CMP3Info::GetMPEGVersionLayer(TCHAR *mpeglayer)
{
	if (m_CoreMP3Info.mpeg25)
		_stprintf(mpeglayer, _T("MPEG 2.5, layer %d"), m_CoreMP3Info.lay);
	else
		_stprintf(mpeglayer, _T("MPEG %d, layer %d"), m_CoreMP3Info.lsf + 1, m_CoreMP3Info.lay);

}

