/*****************************************************************************
    This file is part of MP3ext.
    Copyright (C) 1999  Michael Mutschler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    email: michael@mutschler.de
******************************************************************************/


#include "priv.h"

#include <sys/stat.h>
#include "StringConverter.h"


#ifndef ODS
#ifdef _DEBUG
#define ODS(sz) OutputDebugStringA(sz)
#else
#define ODS(sz)
#endif
#endif

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#endif


#ifdef __cplusplus
// ***********************************************************
string *wstr2str(const wstring& wstr, string& str)
{
	int wstrLength = wstr.length();
	LPSTR cstr = (LPSTR)alloca(wstrLength * sizeof(char));
	if ((cstr != NULL) &&
		 WideCharToMultiByte(CP_ACP, 0, wstr.c_str(), wstrLength,
									cstr, wstrLength, NULL, NULL) )
	{
		str.assign(cstr, wstrLength);
	}
	else
		str = "";
	return(&str);
}

// ***********************************************************
string *wstr2str(const wstring& wstr)
{
	string *str = new string();
	return(wstr2str(wstr, *str));
}

// ***********************************************************
wstring *str2wstr(const string& str, wstring& wstr)
{
	int strLength = str.length();
	LPWSTR wcstr = (LPWSTR)alloca(strLength * sizeof(wchar_t));
	if ((wcstr != NULL) &&
		 MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, str.c_str(), strLength,
									wcstr, strLength) )
	{
		wstr.assign(wcstr, strLength);
	}
	else
		wstr = L"";
	return(&wstr);
}

// ***********************************************************
wstring *str2wstr(const string&   str)
{
	wstring *wstr = new wstring();
	return(str2wstr(str, *wstr));
}

// ***********************************************************
// check if the provided tstring can be converted to a simple string
// returns ID3V2_UNICODE if it fails, or ID3V2_ANSI if successful
bool canUseAnsi(tstring& tstr)
{
#ifdef UNICODE
	BOOL conversionError;
	WideCharToMultiByte(CP_ACP, 0, tstr.c_str(), -1, NULL, 0, NULL, &conversionError);
	return(!conversionError);
#else
	return true;
#endif
}

/** ***********************************************************
 * Load a file into a string. This method takes care if the file
 * is ASCII or UNICODE, and does the proper conversion.
 */
void loadString(tstring& str, LPCTSTR file)
{
	str.resize(0);

	struct _stat sinfo;
	int err = _tstat(file, &sinfo);
	if (err == 0)	// on error, do nothing
	{
		/* get the size of the file */
		int flen = sinfo.st_size;
		LPSTR buffer=NULL;
		// load the file
		if ((buffer = (LPSTR)malloc(flen+2)) != NULL) {
			FILE *fh;
			if ((fh = _tfopen(file, _T("rb"))) != NULL) {
				int bytes = fread(buffer, 1, flen, fh);
				fclose(fh);
				// set the terminating null-char
				buffer[flen]=buffer[flen+1]='\0';
				// check if we have unicode...
				LPWSTR wbuf = (LPWSTR)buffer;
				if (*wbuf == 0xFEFF) {	// I can read this directly
					w2tcpy(str, &wbuf[1]);
				}
				else if (*wbuf == 0xFFFE) {	// wrong byte-order :-(
					LPWSTR wstr = &wbuf[1];
					WCHAR w;
					while ((w=*wstr) != '\0') {
						// swap w-bytes, and write it back
						*wstr++ = (w>>8) + (w&0xFF<<8);
					}
					w2tcpy(str, &wbuf[1]);
				}
				else {
					a2tcpy(str, buffer);
				}

			} else
				ODS("Failed to open file\n");
			free(buffer);
		} else
			ODS("Failed to allocate mem for reading\n");
	} else
		ODS("Error finding file\n");
}


/** ******************************************************************
 * Save a string into a file. This method saves as ASCII when possible.
 */
void saveString(tstring& str, LPCTSTR file)
{
	int buflen = 0;
	void *buffer=NULL;
	// check what we can write
	if (canUseAnsi(str)) {
		// write ANSI-file
		if (buffer = malloc(str.length()+1)) {
			t2acpy((LPSTR)buffer, str);
			buflen = strlen((LPSTR)buffer);
		}
	} else {
		if (buffer = malloc((str.length()+1)*sizeof(WCHAR))) {
			t2wcpy((LPWSTR)buffer, str);
			buflen = wcslen((LPWSTR)buffer);
		}
	}
	// save the file
	if (buffer) {
		FILE *fh;
		if ((fh = _tfopen(file, _T("w"))) != NULL) {
			fwrite(buffer, 1, buflen, fh);
			fclose(fh);
		}
		free(buffer);
	}
	else ODS("Error writing tstring in file");
}

/* *********************************************************** */ 
void copyStringToClipboard(tstring& str)
{
	HGLOBAL hmem=NULL;	// pointer to the heap-object
	LPSTR mem = NULL;
	int size = str.length();
	if (((hmem=GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE, size+1)) != NULL) &&
		 (mem=(char *)GlobalLock(hmem)))
	{
		// copy the result to the clipboard, as ANSI
		t2acpy(mem, str);
		//  Unlock, and then open the Clipboard
		GlobalUnlock(hmem); mem = NULL;
		if (OpenClipboard(NULL))
		{
			//Copy the data to the clipboard
			EmptyClipboard();
			if (SetClipboardData(CF_TEXT, hmem)) hmem=NULL;

			// Close the clipboard
			CloseClipboard();
		}
	}
	else
	{	// error opening the target
		MessageBeep(MB_ICONEXCLAMATION);
		MessageBox(NULL, TEXT_CLIPBORADMEMERROR,
					  NULL, MB_OK|MB_ICONERROR);
	}

	if (mem)  GlobalUnlock(hmem);
	if (hmem) GlobalFree(hmem);
}




#endif //__cplusplus



